# DegreeCox Paper results
#        _                                            _              _
#       | |                                          | |            | |
#     __| | ___  __ _ _ __ ___  ___    __ _ _ __   __| |  _ __   ___| |_ ___ _____  __
#    / _` |/ _ \/ _` | '__/ _ \/ _ \  / _` | '_ \ / _` | | '_ \ / _ \ __/ __/ _ \ \/ /
#   | (_| |  __/ (_| | | |  __/  __/ | (_| | | | | (_| | | | | |  __/ || (_| (_) >  <
#    \__,_|\___|\__, |_|  \___|\___|  \__,_|_| |_|\__,_| |_| |_|\___|\__\___\___/_/\_\
#                __/ |
#               |___/
rm(list=ls())

# Loading necessary libraries

# Survival coxph functions
library(survival)

# Nice plots of survival curves
library(ggfortify)

# Mapvalues and other functions
library(plyr)

# Plot grid of ggplots
library(gridExtra)

# Parallel
library(doMC) # parallel computing
library(foreach)
registerDoMC(cores=4)

# Logger (instead of prints)
library(futile.logger)
flog.appender(appender.tee('logger.txt'))

#
#
# How to install verissimo library (it is not at cran):
#
# library(devtools)
# install_github("averissimo/verissimo")
# 
# OR
#
# download :
#  - https://github.com/averissimo/verissimo-rpackage/archive/v1.0.tar.gz
# and run :
# 
# install.packages("path/to/downloaded/source", repos = NULL)
#
library(verissimo)

#                          __                  _   _                 
#                         / _|                | | (_)                
#     __ _ _   ___  __   | |_ _   _ _ __   ___| |_ _  ___  _ __  ___ 
#    / _` | | | \ \/ /   |  _| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
#   | (_| | |_| |>  <    | | | |_| | | | | (__| |_| | (_) | | | \__ \
#    \__,_|\__,_/_/\_\   |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
#                                                                    

# how to convert different time data to months
time2months <- list(tothill = 1 / 365 * 12, bono = 12, tcga = 1 / 365 * 12)

# calc rmse from residuals
calc_rmse <- function(my.residuals) {
  return(sqrt(mean(my.residuals^2)))
}

# print important metrics
print.metrics <- function(name, rmse, len.rmse, c.index, log.rank) {
  if (file.exists(out.csv.filename))
    df.read <- read.csv2(out.csv.filename, row.names = 1)
  else
    df.read <- data.frame()
  # custom format
  my.format <- function(num) { format(num, scientific = T, nsmall = 8) }
  #
  df.add <- data.frame(rmse     = my.format(rmse), 
                       c.index  = my.format(c.index), 
                       log.rank = my.format(log.rank), 
                       len.rmse = len.rmse,
                       row.names = name)
  #
  write.csv2(file = out.csv.filename, x = rbind(df.read, df.add))
  #
  flog.info('    %s\t%s\t%s\t%s\t%d', 
            name, 
            my.format(rmse),
            my.format(c.index),
            my.format(log.rank),
            len.rmse)
  
}
# get concordance index
get.c_index <- function(concordance.data) {
  return(concordance.data[["concordant"]]/ (concordance.data[["concordant"]] + concordance.data[["discordant"]]))
}
#
top.ranked           <- NULL
#
result.plots <- list()

#                _   _                 
#               | | (_)                
#     ___  _ __ | |_ _  ___  _ __  ___ 
#    / _ \| '_ \| __| |/ _ \| '_ \/ __|
#   | (_) | |_) | |_| | (_) | | | \__ \
#    \___/| .__/ \__|_|\___/|_| |_|___/
#         | |                          
#         |_|
#
my.options <- list()
# name across matlab and R to determine which results to use
my.options$result.set     <- 'tothill_many'
# networks that should be used
my.options$all.networks   <- c('funlinkage', 'correlation')
# pretty name of networks
my.options$network.pretty <- list(funlinkage = 'Functional Map', correlation = 'Correlation')
# pretty name of datasets
my.options$datasets.pretty <- list(tcga = 'TCGA', bono = 'Bonome', tothill = 'Tothill')
# training sets that should be used
my.options$all.types      <- c('tcga', 'bono', 'tothill')
# which should be the test sets used in each type
my.options$all.target     <- list(tcga = c('bono','tothill','tcga'), bono = c('tcga', 'tothill','bono'), tothill = c('bono', 'tcga','tothill'))
# how to separate survival groups in low and high risk patients
my.options$probs_sep      <- c(.4, .6)
# use normalized data (should be used if trained with normalized data)
my.options$use.normalized.xdata <- TRUE
# type of residuals that should be used
my.options$residuals.type <- 'deviance' # deviance, martingale, score, schoenfeld, dfbeta, dfbetas, scaledsch, partial

#
# custom settings for faster runs
# my.options$all.types  <- c('tcga')
# my.options$all.networks <- 'correlation'
# my.options$all.networks <- 'funlinkage'
# my.options$all.target <- list(tcga = c('bono','tothill'))
# my.options$all.target <- list(tcga = c('tcga'))
#
#                _   _         
#               | | | |        
#    _ __   __ _| |_| |__  ___ 
#   | '_ \ / _` | __| '_ \/ __|
#   | |_) | (_| | |_| | | \__ \
#   | .__/ \__,_|\__|_| |_|___/
#   | |                        
#   |_|                        
# base directories for data retrieval and output
basedir.data <- file.path('data', 'results', my.options$result.set, 'determine_btas')
basedir.out  <- file.path('data', 'imgs')

# output csv file
out.csv.filename <- file.path(basedir.out, 'metrics.csv')
if (file.exists(out.csv.filename))
  file.remove(out.csv.filename)

#                     
#                     
#    _ __ _   _ _ __  
#   | '__| | | | '_ \ 
#   | |  | |_| | | | |
#   |_|   \__,_|_| |_|
#                     
#    
#
call.me <- function() {
  for (network.name in my.options$all.networks) { # network.name <- my.options$all.networks[1]
    result.plots[[network.name]] <- list()
    #
  
    # keep values from initial guess
    #  which should be the coefficients from netcox / degree cox
    my.control <- coxph.control(iter.max = 0)
  
    #
    # Run for data from different sources
    #
    #for (my.type in c('tcga','tothill','bono')) { # my.type <- 'tcga'
    for (my.type in my.options$all.types) { # my.type <- my.options$all.types[1]
      result.plots[[network.name]][[my.type]] <- list()
    #for (my.type in c('bono', 'tcga')) {
      #
      if (is.null(my.options$all.target)) {
        my.all.target <- list()
        my.all.target[[my.type]] <- my.type
      } else {
        my.all.target <- my.options$all.target
      }
  
      for (target in my.all.target[[my.type]]) { # target <- my.all.target[[my.type]][1]
        result.plots[[network.name]][[my.type]][[target]] <- list()
        if (target != my.type) {
          # load xdata
          load(file = paste0(basedir.data, '/out_common_', network.name,'_', target, '_01k_res_01g.RData'))
          target.xdata <- xdata
          # use normalized xdata, as the coefficients were calculated on this data
          if (my.options$use.normalized.xdata) {
            target.xdata <- xdata_n
            target.ydata <- ydata
            target.genes <- gene.names
          }
          # remove data from environment
          rm(xdata, xdata_n, ydata, gene.btas.dc, gene.btas.nc, gene.names)
        }
  
        #
        # load data imported from matlab
        load(file = paste0(basedir.data, '/out_common_', network.name, '_', my.type, '_01k_res_01g.RData'))
        #
        if (target == my.type) {
          my.target <- my.type
          # use normalized xdata, as the coefficients were calculated on this data
          if (my.options$use.normalized.xdata)
            xdata <- xdata_n
        } else {
          my.target <- target
          if (!all(target.genes %in% gene.names) )
            stop('Not all genes are the same between data sets!')
          xdata <- target.xdata[,gene.names]
          ydata <- target.ydata
        }
        #
        flog.info(paste0('Running for network ', network.name,' data trained by ', my.type, ' tested on ', my.target))
        #
        #
        coxph.model.file <- paste0(basedir.data, '/', network.name, '_coxph_', my.type, '_with_xdata_', my.target, '_01k_01g.RData')
        top.pvalues.file <- paste0(basedir.data, '/', network.name, '_top_pvalues_', my.type, '_with_xdata_', my.target, '_', format(top.ranked), '.RData')
        #
        #
        # create dataframe for coxph usage
        new.df <- data.frame( time = ydata$time, status = ydata$status)
        new.df <- cbind(new.df, xdata)
        if (!file.exists(coxph.model.file)) {
          # Simulate Cox model
          coxph.model <- list()
          coxph.model$data   <- new.df
          flog.info('  Generating coxph models:')
          #
          get.coxph <- function(btas) {
            return(tryCatch(coxph.fit(xdata.matrix, Surv(coxph.model$data$time, coxph.model$data$status),
                                      init = btas, control = my.control,
                                      strata = NULL, method = 'breslow', rownames = NULL)
                            , error = function(e) {
                              flog.info('Error when calling coxph.fit.')
                              retrun(coxph(Surv(time, status) ~ . , data = coxph.model$data,
                                    init = btas,
                                    control = my.control, ties = 'breslow'))
                            }))
          }
          #
  
          #
          input.args <- list(degree = gene.btas.dc,
                             netcox = gene.btas.nc,
                             lasso  = gene.btas.la,
                             ridge  = gene.btas.ri)
          #
          xdata.matrix <- as.matrix(xdata)
          res = foreach ( iaix = 1:length(input.args)) %dopar% {
            flog.info('    %s - %s - %s - %s..' , network.name, my.type, my.target, names(input.args)[iaix])
            coxph.model[[names(input.args)[iaix]]] <- get.coxph(input.args[[iaix]])
          }
          flog.info('Finished with generating cox models')
          for (iaix in 1:length(input.args)) {
            coxph.model[[names(input.args)[iaix]]] <- res[[iaix]]
          }
          #
          save(coxph.model, file = coxph.model.file)
          load(coxph.model.file)
        } else {
          flog.info('  Loading %s', coxph.model.file)
          if (exists('coxph.model'))
            rm('coxph.model')
          load(coxph.model.file)
        }
  
        if (!is.null(top.ranked)) {
          if (!file.exists(top.pvalues.file)) {
            degree <- list()
            netcox <- list()
            #
            #
            # DegreeCox
            #
            degree$top.ranked <- top.ranked
            degree$gene.list  <- names(sort(abs(gene.btas.dc))[1:top.ranked])
            degree$result     <- run_single_variables(favorite_genes = degree$gene.list,
                                                      xdata   = xdata,
                                                      ydata   = ydata,
                                                      prefix  = 'DegreeCox',
                                                      run_number = 1,
                                                      plot_indiv = FALSE,
                                                      plot_p_values = FALSE)
  
            #
            #
            # NetCox
            #
            netcox$top.ranked <- top.ranked
            netcox$gene.list  <- names(sort(abs(gene.btas.nc))[1:top.ranked])
            netcox$result     <- run_single_variables(favorite_genes = netcox$gene.list,
                                                      xdata   = xdata,
                                                      ydata   = ydata,
                                                      prefix  = 'NetCox',
                                                      run_number = 1,
                                                      plot_indiv = FALSE,
                                                      plot_p_values = FALSE)
            #
            degree$name <- 'DegreeCox'
            netcox$name <- 'NetCox'
            #
            save(netcox, degree, file = top.pvalues.file)
          } else {
            flog.info(paste0('  Loading ', top.pvalues.file))
            load(top.pvalues.file)
          }
        } else {
          netcox <- list()
          degree <- list()
          lasso <- list()
          ridge <- list()
        }
        #
        #
        flog.info('  Calculate residuals')
        degree$residuals <- residuals(coxph.model$degree, type = my.options$residuals.type)
        netcox$residuals <- residuals(coxph.model$netcox, type = my.options$residuals.type)
        lasso$residuals  <- residuals(coxph.model$lasso,  type = my.options$residuals.type)
        ridge$residuals  <- residuals(coxph.model$ridge,  type = my.options$residuals.type)
        ridge$residuals  <- ridge$residuals[ridge$residuals!=Inf]
        #
        # TODO: remove after re-run everything
        degree$name <- 'DegreeCox'
        netcox$name <- 'Net-Cox'
        lasso$name  <- 'Lasso'
        ridge$name  <- 'Ridge'
        #
        # plot residues
        flog.info('  Plotting residuals')
        if (my.type == 'bono') {
          my.type.name <- 'Bonome'
        } else if (my.type == 'tcga') {
          my.type.name <- 'TCGA'
        } else {
          my.type.name <- 'Tothill'
        }
        my.xlim <- c(min(density(ridge$residuals)$x, density(lasso$residuals)$x, density(degree$residuals)$x, density(netcox$residuals)$x)-1,
                     max(density(ridge$residuals)$x, density(lasso$residuals)$x, density(degree$residuals)$x, density(netcox$residuals)$x)+1)
        my.xlim[1] <- -5
        my.ylim <- c(0,.95)
        if (my.target == 'bono') {
          my.target.name <- 'Bonome'
        } else if(my.target == 'tcga') {
          my.target.name <- 'TCGA'
        } else if(my.target == 'tothill') {
          my.type.name <- 'Tothill'
        }
        filename <- paste0(network.name, '_', my.options$residuals.type, '_residues',
                           '_', tolower(my.type.name),
                           '_with_xdata_', tolower(my.target.name))
        my.plot.residuals(list(degree$residuals, netcox$residuals, ridge$residuals, lasso$residuals),
                         paste0(c('DegreeCox', '      NetCox ','          Ridge', '          Lasso')), my.options$residuals.type,
                         my.xlim = my.xlim, my.ylim = my.ylim,
                         filename = filename, base.directory = file.path(basedir.out, 'residuals'))
        title(paste0('Trained on ', my.type, ' and tested on ', my.target))
  
        if (!is.null(top.ranked)) {
          for (my.data in list(degree, netcox)) {
            plot_single_gene(my.data$result$p_values$values,
                             my.data$result$p_values$title,
                             xlabs  = my.data$result$p_values$xlabs,
                             type   = my.data$result$p_values$type,
                             prefix = my.data$result$p_values$prefix,
                             cex.axis = .7,
                             filename = tolower(paste(network.name,
                                                my.data$name,
                                                my.type.name,
                                                'with_xdata', my.target.name,
                                                'top', format(length(my.data$result$p_values$xlabs))
                                                , sep = '_')))
            flog.info(paste0(network.name, ': Trained with ', my.type.name, ' and tested with ', my.target.name, ' the # of p-values under 0.05 in ', my.data$name, ': ', sum(my.data$result$p_values$values < 0.05)))
          }
        }
  
        #
        common_genes <- names(gene.btas.dc[degree$gene.list]) %in% names(gene.btas.nc[netcox$gene.list])
        common_genes <- names(gene.btas.dc[degree$gene.list])[common_genes]
        #
        flog.info('  Converting days / years to months')
        ydata_days <- ydata
        ydata_days$time <- ydata_days$time * time2months[[target]]
  
        flog.info('  Creating KM curve plots (see result.plot for all plots)')
        surv_names <- function(prefix, suffix) {
          paste0(type, '-')
        }
        lasso.logrank  <- draw.kaplan(chosen.btas = list(Lasso = gene.btas.la),     xdata = xdata, ydata = ydata_days, probs = my.options$probs_sep)
        ridge.logrank  <- draw.kaplan(chosen.btas = list(Ridge = gene.btas.ri),     xdata = xdata, ydata = ydata_days, probs = my.options$probs_sep)
        netcox.logrank <- draw.kaplan(chosen.btas = list(NetCox = gene.btas.nc),    xdata = xdata, ydata = ydata_days, probs = my.options$probs_sep)
        degree.logrank <- draw.kaplan(chosen.btas = list(DegreeCox = gene.btas.dc), xdata = xdata, ydata = ydata_days, probs = my.options$probs_sep)
        #
        result.plots[[network.name]][[my.type]][[target]]$lasso  <- list(call = list( Lasso = gene.btas.la), xdata = xdata, ydata = ydata_days)
        result.plots[[network.name]][[my.type]][[target]]$ridge  <- list(call = list( Ridge = gene.btas.ri), xdata = xdata, ydata = ydata_days)
        result.plots[[network.name]][[my.type]][[target]]$netcox <- list(call = list(NetCox = gene.btas.nc), xdata = xdata, ydata = ydata_days)
        result.plots[[network.name]][[my.type]][[target]]$degree <- list(call = list(Degree = gene.btas.dc), xdata = xdata, ydata = ydata_days)
        #
        flog.info('  Printing metrics: rmse / c-index / logrank p-value test / size of dataset')
        str.prefix <- paste0(network.name, '  ', my.type, '  ', my.target, '  ')
        print.metrics(paste0(str.prefix, 'netcox'), calc_rmse(netcox$residuals), length(netcox$residuals), get.c_index(coxph.model$netcox$concordance), netcox.logrank$pvalue)
        print.metrics(paste0(str.prefix, 'degree'), calc_rmse(degree$residuals), length(degree$residuals), get.c_index(coxph.model$degree$concordance), degree.logrank$pvalue)
        print.metrics(paste0(str.prefix, 'lasso'), calc_rmse(lasso$residuals), length(lasso$residuals), get.c_index(coxph.model$lasso$concordance), lasso.logrank$pvalue)
        print.metrics(paste0(str.prefix, 'ridge'), calc_rmse(ridge$residuals), length(ridge$residuals), get.c_index(coxph.model$ridge$concordance), ridge.logrank$pvalue)
      }
    }
  }
  return(result.plots)
}
#
flog.info("Calculating all necessary data")
result.plots <- call.me()
flog.info("")
flog.info("")
flog.info("")
flog.info("Plotting all survival curves")


#                         _            _                                    
#                        (_)          | |                                   
#    ___ _   _ _ ____   _____   ____ _| |     ___ _   _ _ ____   _____  ___ 
#   / __| | | | '__\ \ / / \ \ / / _` | |    / __| | | | '__\ \ / / _ \/ __|
#   \__ \ |_| | |   \ V /| |\ V / (_| | |   | (__| |_| | |   \ V /  __/\__ \
#   |___/\__,_|_|    \_/ |_| \_/ \__,_|_|    \___|\__,_|_|    \_/ \___||___/
#                                                                           
#                                        
#my.plot.call <- function() {
  for (typeix in 1:length(my.options$all.types)) {
    my.type <- my.options$all.types[typeix]
    for (targetix in 1:length(my.options$all.target[[my.type]])) {
  
      my.target <- my.options$all.target[[my.type]][targetix]
  
      gen.names <- function(my.reg, my.network = NULL) {
        str <- paste0(my.options$datasets.pretty[[my.type]], '_tested_with_', my.options$datasets.pretty[[my.target]])
        if (!is.null(my.network))
          str <- paste0(str, '_', my.network)
        str <- paste0(str, '_', my.reg)
        return(str)
      }
      # flog.info("  Trained with %s \t Tested on %s", my.type, my.target)
      #
      temp.all.la <- result.plots$funlinkage[[my.type]][[my.target]]$lasso
      ydata <- result.plots$funlinkage[[my.type]][[my.target]]$lasso$ydata
      xdata <- result.plots$funlinkage[[my.type]][[my.target]]$lasso$xdata
      #
      temp.all.ri <- result.plots$funlinkage[[my.type]][[my.target]]$ridge
      temp.fun.nc <- result.plots$funlinkage[[my.type]][[my.target]]$netcox
      temp.cor.nc <- result.plots$correlation[[my.type]][[my.target]]$netcox
      temp.fun.dc <- result.plots$funlinkage[[my.type]][[my.target]]$degree
      temp.cor.dc <- result.plots$correlation[[my.type]][[my.target]]$degree
  
      basedir.out.kaplan <- file.path(basedir.out, 'kaplan')
      draw.kaplan.temp <- function(my.call, filename) {
        return(draw.kaplan(my.call,
                           xdata, ydata, 
                           probs          = my.options$probs_sep, save.plot = T,
                           base.directory = basedir.out.kaplan, filename = filename,
                           ylim           = c(0.04761,1),
                           xlim           = c(10,100)))
      }
      draw.kaplan.temp(temp.fun.dc$call, filename = gen.names('degree', 'funlinkage'))$plot
      lasso.all.logrank  <- draw.kaplan.temp(temp.all.la$call, filename = gen.names('lasso'))$plot
      ridge.all.logrank  <- draw.kaplan.temp(temp.all.ri$call, filename = gen.names('ridge'))$plot
      netcox.fun.logrank <- draw.kaplan.temp(temp.fun.nc$call, filename = gen.names('netcox', 'funlinkage'))$plot
      netcox.cor.logrank <- draw.kaplan.temp(temp.cor.nc$call, filename = gen.names('netcox', 'correlation'))$plot
      degree.fun.logrank <- draw.kaplan.temp(temp.fun.dc$call, filename = gen.names('degree', 'funlinkage'))$plot
      degree.cor.logrank <- draw.kaplan.temp(temp.cor.dc$call, filename = gen.names('degree', 'correlation'))$plot
      plot.fun <- draw.kaplan.temp(c(temp.fun.nc$call, temp.fun.dc$call, temp.all.la$call, temp.all.ri$call),
                              filename = paste0('All_funlinkage_',my.type, '_', my.target))$plot
  
      plot.cor <- draw.kaplan.temp(c(temp.cor.nc$call, temp.cor.dc$call, temp.all.la$call, temp.all.ri$call),
                              filename = paste0('All_correlation_',my.type, '_', my.target))$plot
      my.grid <- grid.arrange(lasso.all.logrank, ridge.all.logrank, netcox.fun.logrank, netcox.cor.logrank, degree.fun.logrank, degree.cor.logrank,
                   ncol = 2)
      my.save.ggplot(paste0('aux_', my.type, '_', my.target), my.plot = my.grid, base.directory = basedir.out.kaplan, width = 14, height = 9.5 )
      
    }
  }
#}
my.plot.call()

# free up memory
# rm('result.plots')

